/* CubeCell Plus LoRa Pump Controller with Command ACK
  
   Function:
   LoRa Pump Controller - CubeCellPlusPromReceiverEcho but now using formal ACK-Packet Echo response
  
   Description:
   1. Uses packet structures defined in the PacketHandler library
   2. Test sketch for [Pump] ACK protocol development
  
   30 Apr 2023 v0.0.4  Packet Echo with ACK Flag
   16 May 2023 v0.0.5  REL flag with interrupt for local override
               v0.0.6  Add REL/ACK buffer
   22 May 2023 v0.0.7  TX_State code rearrangement
   30 May 2023 v0.0.8  Remove reference to EEPROM Handler
    1 Jun 2023 v0.0.9  General housekeeping to get the various implementations [more or less] consistent
    1 Jun 2023 v0.0.10 Change 'local override' interrupt to GPIO11 [as per 10068-BHCP v2.3+]
    3 Jun 2023 v0.0.11 Report pump status with sensor readings
   26 Jun 2023 v0.0.12 Add watchdog timer
    6 Jul 2023 v0.0.13 Correct checkAckBuffer logic
    9 Jul 2023 v0.0.14 Randomise reportingPeriod to avoid Node 'update syncing'
   
   9 Jul 2023
   Digital Concepts
   www.digitalconcepts.net.au
 */

#include <innerWdt.h>

#include <Arduino.h>
#include <Wire.h>               // Need this to define I2C bus
#include <LoRaWan_APP.h>
#include <HT_SH1107Wire.h>     // CubeCell Plus display library
#include <Seeed_BME280.h>

#include <LangloLoRa.h>         // LoRa configuration parameters
#include <PacketHandler.h>      // [Langlo] LoRa Packet management
#include <LinkedList.h>         // Linked List support for REL/ACK buffer
#include <CubeCellPlusPins.h>   // Heltec CubeCell Plus pin definitions
#include <nodeP1PCConfig.h>     // Specific local Node configuration parameters

/*
 * set LoraWan_RGB to 1,the RGB active in loraWan
 * RGB red means sending;
 * RGB green means received done;
 */
#ifndef LoraWan_RGB
#define LoraWan_RGB 0
#endif

//const int pump_Local_Override = USER_KEY;  // GPIO16 - Seems to generate a lot of false hits.
#define pump_Local_Override GPIO11
#define pump_Ctrl           GPIO5

struct softwareRevision {
  const uint8_t major;			// Major feature release
  const uint8_t minor;			// Minor feature enhancement release
  const uint8_t minimus;		// 'Bug fix' release
};

softwareRevision sketchRevision = {0,0,14};
bool softRev = true;

// The reporting period is randomised to help avoid syncing of Node updates
// Even if near synchronous updates from different Nodes don't collide,
// the Gateway cannot process multiple packets in quick succession

const int lowerReportingLimit = 55000; // 55 sec
const int upperReportingLimit = 65000; // 65 sec
int reportingPeriod; // Will be randomised between lowerReportingLimit and upperReportingLimit

uint16_t messageCounter = 0;
const int wdtInterval = 60;  // Seconds
unsigned long wdtLimit = 180000;  // Milliseconds
unsigned long lastMessageTime = 0;
bool pumpTimerActive = false;
bool turnPumpOff = true;

const int oledTimeout = 30000; // Off after 30 sec
bool oledActive = true;
bool oledOn = true;

const int resendInterval = 1;
const int eventCycleTime = resendInterval * 1000;
const int resendLimit = 3;
int resendCount = 0;
bool ackReceived = false;
bool resendTimerActive = false;

struct bufferElement {
  PacketHandler packet;
  unsigned long startTime;
  int resendCount;
};
bufferElement element;
LinkedList<bufferElement> *ackBuffer = new LinkedList<bufferElement>();
int bufferSize;
bool checkAckBuffer = false;

bool ackRequired = true;
bool sendAck = false;
bool readSensors = false;

bool changePumpState = false;
uint8_t pumpId = 2;
PH_powerStatus pumpPowerState = PH_POWER_OFF;
PH_packetType ackPacketType, messageType;
PacketHandler inPacket, outPacket;

// Sensor variables
uint16_t humidity = 0;
uint16_t pressure = 0;
uint16_t temperature = 0;
BME280 bme280;

// OLED Display
SH1107Wire  myDisplay(0x3c, 500000, SDA, SCL, GEOMETRY_128_64, GPIO10), *displayPtr; // addr, freq, sda, scl, resolution, rst

// Timers
static TimerEvent_t oledTimer;
static TimerEvent_t resendTimer;
static TimerEvent_t sensorTimer;
static TimerEvent_t pumpWatchDogTimer;

static RadioEvents_t RadioEvents;
bool lora_idle = true;
void OnTxDone( void );
void OnTxTimeout( void );
void OnRxDone( void );

typedef enum {
  RX_State,
  TX_State
} States_t;

States_t state;
int16_t rssi,rxSize;

void setup() {
  boardInitMcu( );
  reportingPeriod = random(lowerReportingLimit,upperReportingLimit);
  Serial.begin(115200);
  delay(500);
  Serial.print("[setup] ");
  Serial.print(myName);
  Serial.println(" CubeCell Plus Pump Controller Test");
  
  Serial.println("[setup] Commence Set-up...");

  pinMode(Vext,OUTPUT);
  digitalWrite(Vext,LOW);     // Turn on the external power supply

  pinMode(pump_Ctrl,OUTPUT);
  digitalWrite(pump_Ctrl,LOW);     // Turn the pump OFF

  // Interrupt to change pump state

  PINMODE_INPUT_PULLUP(pump_Local_Override);
//  pinMode( pump_Local_Override,INPUT );
  attachInterrupt( pump_Local_Override, pumpStateControl, FALLING );

  // Interrupt to activate OLED display
/*    
  PINMODE_INPUT_PULLUP(P3_3);  // P3_3 == USER_KEY
  attachInterrupt(P3_3,toggleDisplay,FALLING);
*/                                               

  displayPtr = &myDisplay;
//  myDisplay.displayOn(); // Switch on display
//  delay(10);
//  oledOn = true;
  myDisplay.init();
  myDisplay.setFont(ArialMT_Plain_10);
  myDisplay.screenRotate(ANGLE_180_DEGREE);
    
  myDisplay.clear();
  myDisplay.setFont(ArialMT_Plain_10);
  myDisplay.setTextAlignment(TEXT_ALIGN_LEFT);
  myDisplay.drawString(5,5,"LoRa Pump Controller");
  myDisplay.display();
  delay(1000);

  if ( softRev ) {
    Serial.println("[setup] softRev = true");
    Serial.println("[setup] Check display");
    displaySoftRev();
    delay(1000);
    softRev = false;
  }

//  Initialise LoRa

  Serial.println("[setup] Initialise LoRa...");

  RadioEvents.TxDone = OnTxDone;
  RadioEvents.TxTimeout = OnTxTimeout;
  RadioEvents.RxDone = OnRxDone;
  Radio.Init( &RadioEvents );
  
  Radio.SetChannel( RF_FREQUENCY );
  Radio.SetTxConfig( MODEM_LORA, TX_OUTPUT_POWER, 0, LORA_BANDWIDTH,
                                  LORA_SPREADING_FACTOR, LORA_CODINGRATE,
                                  LORA_PREAMBLE_LENGTH, LORA_FIX_LENGTH_PAYLOAD_ON,
                                  true, 0, 0, LORA_IQ_INVERSION_ON, 3000 );
  Radio.SetRxConfig( MODEM_LORA, LORA_BANDWIDTH, LORA_SPREADING_FACTOR,
                                  LORA_CODINGRATE, 0, LORA_PREAMBLE_LENGTH,
                                  LORA_SYMBOL_TIMEOUT, LORA_FIX_LENGTH_PAYLOAD_ON,
                                  0, true, 0, 0, LORA_IQ_INVERSION_ON, true );
  Radio.Sleep( );
  rssi=0;

  Serial.println("[setup] LoRa initialisation complete");
  Serial.println();
  
  myDisplay.clear();
  myDisplay.drawString(5,20,"LoRa Initialisation OK");
  myDisplay.display();
  delay(1000);

  // initialise the Packet object instances
  
  inPacket.begin();
  messageCounter++;
  outPacket.begin(gatewayMAC,myMAC,messageCounter);

  // and set up the pointers to the OLED display (to pass to the Packet object when required)

  displayPtr = &myDisplay;

// OLED Deactivation Timer

  TimerInit( &oledTimer, deactivateOled );
  Serial.println("[setup] OLED deactivation timer started...");
  TimerSetValue( &oledTimer, oledTimeout );
  TimerStart( &oledTimer );

 // Sensor Read Timer

  TimerInit( &sensorTimer, setSensorFlag );
  Serial.println("[setup] Sensor read timer started...");
  TimerSetValue( &sensorTimer, reportingPeriod );
  TimerStart( &sensorTimer );

// Buffer Processing Timer

  TimerInit( &resendTimer, resendFunction );
  Serial.println("[setup] ACK timer initialised");
  TimerSetValue( &resendTimer, eventCycleTime );
 
// Pump Watch Dog Timer

  TimerInit( &pumpWatchDogTimer, pumpStateCheck );
  Serial.println("[setup] Pump timer initialised");
  TimerSetValue( &pumpWatchDogTimer, wdtLimit );

 /* Enable the WDT. 
  * autoFeed = false: do not auto feed wdt.
  * autoFeed = true : it auto feed the wdt in every watchdog interrupt.
  */
//  innerWdtEnable(autoFeed);
 
   // Let the broker know there's been a reset
  
  outPacket.setPacketType(RESET);
  outPacket.setResetCode(0);
  outPacket.serialOut();
  turnOnRGB(COLOR_SEND,0);
  sendMessage(outPacket);
  turnOffRGB();
  // We need to wait a tick here to let the packet go out before doing anything more with the radio or the packet gets lost
  delay(100);
  messageCounter++;

  // Initialise pump status

  element.packet.begin(gatewayMAC,myMAC,messageCounter,PUMP);
  element.packet.setRelFlag();
  element.packet.setPumpId(pumpId);
  element.packet.setPumpState(PH_POWER_OFF);
  element.packet.serialOut();
  // Add the packet to the buffer of packets awaiting ACK
  element.resendCount = 1;
  element.startTime = millis();
  ackBuffer->add(element);
  // Send the packet and go wait for the acknowledgement
  turnOnRGB(COLOR_SEND,0);
  sendMessage(element.packet);
  turnOffRGB();
  delay(100);
  messageCounter++;
  state = RX_State;

  Serial.println("[setup] Set-up Complete");
  Serial.println();
}

void loop() {
	switch ( state ) {
		case TX_State: {
//      Serial.println("[loop TX] Into TX Mode...");

      // We have four possible tasks to perform
      // 1. Check to see if we need to acknowledge receipt of a packet. We need to do this first because the sender
      //    is waiting for our response;
      // 2. If we have sensor data to send, we need to send that;
      // 3. If a local pump state change has been triggered, we need to advertise the fact to the Gateway/MQTT broker.
      //    This is time-sensitive because we will be expecting an acknowledgement almost immediately. This conflicts with
      //    the similar requirements of ACK buffer processing (4.), so we can only perform one of these functions in any one cycle;
      // 4. If we have packets that require acknowledgement, and need to be resent, we send them (only one each cycle) last, because
      //    we need to transition immediately thereafter into RX_State to receive the acknowledgement.

      if ( sendAck ) {
        Serial.println("[loop TX sendAck] Begin...");
        inPacket.setDestinationMAC(inPacket.sourceMAC());
        inPacket.setSourceMAC(myMAC);
        inPacket.setAckFlag();
        inPacket.hexDump();
        inPacket.serialOut();
        if ( oledOn ) {
          inPacket.displayOut(displayPtr);
        }
        turnOnRGB(COLOR_SEND,0);
        sendMessage(inPacket);
        turnOffRGB();
        sendAck = false;
        // We need to wait a tick here to let the packet go out before doing anything more with the radio
        delay(100);
        Serial.println("[loop TX sendAck] End");
      }
      
      if ( readSensors ) {
        Serial.println("[loop TX readSensors] Begin...");
        outPacket.setSequenceNumber(messageCounter);
        readBatteryVoltage();
        outPacket.hexDump();
        outPacket.serialOut();
        turnOnRGB(COLOR_SEND,0);
        sendMessage(outPacket);
        turnOffRGB();
        delay(100);
        messageCounter++;                 // Increment message count

        outPacket.setSequenceNumber(messageCounter);
        readBMESensor();
        outPacket.hexDump();
        outPacket.serialOut();
        turnOnRGB(COLOR_SEND,0);
        sendMessage(outPacket);
        turnOffRGB();
        delay(100);
        messageCounter++;                 // Increment message count

        // Report pump status
        outPacket.setSequenceNumber(messageCounter);
        outPacket.setPacketType(PUMP);
        outPacket.setPumpId(pumpId);
        outPacket.setPumpState(pumpPowerState);
        outPacket.hexDump();
        outPacket.serialOut();
        turnOnRGB(COLOR_SEND,0);
        sendMessage(outPacket);
        turnOffRGB();
        // We need to wait a tick here to let the packet go out before doing anything more with the radio or the packet gets lost
        delay(100);
        messageCounter++;                 // Increment message count

        readSensors = false;
        TimerStart( &sensorTimer );
        Serial.println("[loop TX readSensors] End");
      }

      if ( changePumpState ) {
        Serial.println("[loop TX changePumpState] Begin...");
        bufferElement element;
        element.packet.begin(gatewayMAC,myMAC,messageCounter,PUMP);
        element.packet.setRelFlag();
        element.packet.setPumpId(pumpId);

        if (( pumpPowerState == PH_POWER_ON ) || turnPumpOff ) {
          pumpPowerState = PH_POWER_OFF;
          digitalWrite(pump_Ctrl,LOW);
        } else {
          pumpPowerState = PH_POWER_ON;
          digitalWrite(pump_Ctrl,HIGH);
        }
        
        element.packet.setPumpState(pumpPowerState);
        element.packet.serialOut();
        turnOnRGB(COLOR_SEND,0);
        sendMessage(element.packet);
        turnOffRGB();
        delay(100);
        messageCounter++;                 // Increment message count
        if ( !resendTimerActive ) {
          Serial.println("[loop TX changePumpState] Start resendTimer...");
          TimerStart( &resendTimer );
          resendTimerActive = true;
        }
        // Add the packet to the buffer of packets awaiting ACK
        element.resendCount = 1;
        element.startTime = millis();
        ackBuffer->add(element);
        changePumpState = false;
      } else if ( checkAckBuffer ) {
        Serial.println("[loop TX checkAckBuffer] Begin...");
        // If there is one, check the first element of the ACK buffer
        // We only do one resend per cycle so as not to block a response to the packet
        // that is sent
        bufferSize = ackBuffer->size();
        while ( bufferSize > 0 ) {
          Serial.print("[loop TX checkAckBuffer] ACK buffer non-empty [");
          Serial.print(bufferSize);
          Serial.println("]");

          unsigned long currentTime = millis();
          unsigned long timeInterval;

          Serial.println("[loop TX checkAckBuffer] Retrieve first element...");
          element = ackBuffer->get(0);
          timeInterval = currentTime - element.startTime;
          Serial.print("[loop TX checkAckBuffer] Current Time: ");
          Serial.println(currentTime);
          Serial.print("[loop TX checkAckBuffer] Start Time: ");
          Serial.println(element.startTime);
          Serial.print("[loop TX checkAckBuffer] Time Interval: ");
          Serial.println(timeInterval);
          Serial.print("[loop TX checkAckBuffer] Resend Interval: ");
          Serial.println(eventCycleTime);
          if (timeInterval > eventCycleTime) {
            Serial.println("[loop TX checkAckBuffer] Resend packet... ");
            resendCount = ++element.resendCount;
            Serial.print("[loop TX checkAckBuffer] Resend count: ");
            Serial.println(resendCount);
            if ( resendCount > resendLimit ) {
              Serial.println("[loop TX checkAckBuffer] Resend Limit exceeded... ");
              Serial.println("[loop TX checkAckBuffer] Delete packet...");
              element = ackBuffer->remove(0);
              // May want to do something more at this point...

              bufferSize = ackBuffer->size();
              Serial.print("[loop TX checkAckBuffer] New Buffer Size: ");
              Serial.println(bufferSize);
            } else {
              // Move the element to the end of the queue
              Serial.println("[loop TX checkAckBuffer] Move element to end of queue...");
              element = ackBuffer->remove(0);
              element.startTime = currentTime;
              element.resendCount = resendCount;
              ackBuffer->add(element);
              Serial.println("[loop TX checkAckBuffer] Resending packet...");
              turnOnRGB(COLOR_SEND,0);
              sendMessage(element.packet);
              turnOffRGB();
              delay(100);
              // We need to go see if there's a response to this transmission, so we'll
              // pretend the buffer's empty to get us out of the buffer processing loop
              bufferSize = 0;
            }
          } else {
            bufferSize--;
          }
          Serial.println();
        }
        // We will have exited the loop after resending just one packet, so we need to
        // check to see if there's still more in the buffer that we need to come back for       
        bufferSize = ackBuffer->size();
        if ( bufferSize > 0 ) {
          if ( !resendTimerActive ) {
            Serial.println("[loop TX bufferSize] Start resendTimer...");
            TimerStart( &resendTimer );
            resendTimerActive = true;
          }
        } else {
          checkAckBuffer = false;
        }
        Serial.println("[loop TX checkAckBuffer] End");
      }

//      Serial.println();
      state = RX_State;
      break;
    }
		case RX_State: {
//      Serial.println("[loop RX] Into RX Mode...");
      if ( lora_idle ) {
        lora_idle = false;
        Radio.Rx(0);
      }
      Radio.IrqProcess( );
      state = TX_State;
      break;
		}
    default: {
      state = RX_State;
      break;
    }
	}
}

void OnTxDone( void ) {
//  Serial.println("[OnTxDone] TX done!");
  Radio.Sleep( );
  lora_idle = true;    
  state = RX_State;
}

void OnTxTimeout( void ) {
//  Serial.println("[OnTxTimeout] TX Timeout...");
  Radio.Sleep( );
  lora_idle = true;    
  state = RX_State;
}

void OnRxDone( uint8_t *payload, uint16_t size, int16_t rssi, int8_t snr ){
  turnOnRGB(COLOR_RECEIVED,0);
  inPacket.setContent(payload,size);
  turnOffRGB();
  Radio.Sleep( );
  lora_idle = true;    

  if (inPacket.destinationMAC() == myMAC) {  
/*        
      Serial.println("[OnRxDone] This one's for us...");
*/  
    uint8_t byteCount = inPacket.packetByteCount();
/*
    Serial.println("[OnRxDone] Packet received");
    Serial.println("[OnRxDone] Header content:");
    Serial.print("       DestinationMAC: ");
    Serial.printf("%02X", inPacket.destinationMAC());
    Serial.println();
    Serial.print("            SourceMAC: ");
    Serial.printf("%02X", inPacket.sourceMAC());
    Serial.println();
    Serial.print("           Sequence #: ");
    Serial.println(inPacket.sequenceNumber());
    Serial.print("             REL Flag: ");
    if ( inPacket.relFlag() ) {
      Serial.println("True");
    } else {
      Serial.println("False");
    }
    Serial.print("          Packet Type: ");
    Serial.println(inPacket.packetTypeDescription());
    Serial.print("             ACK Flag: ");
    if ( inPacket.ackFlag() ) {
      Serial.println("True");
    } else {
      Serial.println("False");
    }
    Serial.print("   Payload Byte Count: ");
    Serial.println(inPacket.payloadByteCount());
    Serial.println();
*/
    //  Verify the checksum before we go any further
  
    if (inPacket.verifyPayloadChecksum()) {

      Serial.println("[OnRxDone] Process the packet...");
        
      // Send relevant information to the Serial Monitor

      inPacket.hexDump();
      inPacket.serialOut();
      // print RSSI & SNR of packet
      Serial.print("[OnRxDone] RSSI ");
      Serial.print(rssi);
      Serial.print("   SNR ");
      Serial.println(snr);

    // and finally, respond when required...

      if ( inPacket.ackFlag() ) {
        if ( verifyAck(inPacket) ) {
          // Nothing more to do here at the moment
        }
      } else {
        switch (inPacket.packetType()) {
          case PUMP: {
            pumpPowerState = inPacket.pumpState();
            if ( pumpPowerState == PH_POWER_ON ) {
              digitalWrite(pump_Ctrl,HIGH);
              lastMessageTime = millis();
              if ( pumpTimerActive ) {
                TimerReset( &pumpWatchDogTimer );
              } else {
                TimerStart( &pumpWatchDogTimer );
                pumpTimerActive = true;
              }
            } else {
              digitalWrite(pump_Ctrl,LOW);
              if ( pumpTimerActive ) {
                TimerStop ( &pumpWatchDogTimer );
              }
            }
            break;
          }
          default: {
            Serial.println("[OnRxDone packetType Switch] Well, what am I supposed to do with this...?");
          }
        }
        if ( inPacket.relFlag() ) {
          sendAck = true;
          state = TX_State;
        }
      }

      if ( oledOn ) {
        // Send relevant information to the OLED display
  
        inPacket.displayOut(displayPtr);   
        // print RSSI and SNR of packet
        myDisplay.drawHorizontalLine(0, 52, 128);
        myDisplay.setTextAlignment(TEXT_ALIGN_LEFT);
        myDisplay.drawString(0, 54, "RSSI " + String(rssi) + "dB");
        myDisplay.setTextAlignment(TEXT_ALIGN_RIGHT);
        myDisplay.drawString(128, 54, "SNR " + String(snr) +"dB");
        myDisplay.display();
      }
  
      Serial.println();
      
    } else {
      Serial.println("[OnRxDone] Invalid checksum, packet discarded");
      Serial.println();
      inPacket.erasePacketHeader();
    }
/*
    Serial.println("[receivePacket] Checksum validation disabled");
    Serial.println();
    return true;
*/
  } else {
    
    // No need to go any further
/*
    Serial.print("[OnRxDone] This appears to be for someone else--To: 0x");
    Serial.print(inPacket.destinationMAC(),HEX);
    Serial.print(" From: 0x");
    Serial.println(inPacket.sourceMAC(),HEX);
*/
    inPacket.erasePacketHeader();
  }
}

void sendMessage(PacketHandler packet) {
//  Serial.println("[sendMessage] Finished Packet");
  Radio.Send(packet.byteStream(), packet.packetByteCount());
  Serial.println("[sendMessage] Packet Sent");
  Serial.println();
}

bool verifyAck(PacketHandler ackPacket) {
  // Search the buffer for a matching outstanding ACK
  int i = 0;
  int bufferSize = ackBuffer->size();
  bool found = false;
//  Serial.print("[verifyAck] Buffer size: ");
//  Serial.println(bufferSize);
  while ( i < bufferSize && !found ) {
//    Serial.print("[verifyAck] Checking buffer element: ");
//    Serial.println(i);
    element = ackBuffer->get(i);
    ackPacketType = ackPacket.packetType();
    if ( ackPacket.sourceMAC() == element.packet.destinationMAC() &&
        ackPacket.sequenceNumber() == element.packet.sequenceNumber() && 
        ackPacketType == element.packet.packetType() ) {
      Serial.println("[verifyAck] ACK verified");
      element = ackBuffer->remove(i);  // Crashes if we don't use the form of remove() that copies the deleted element (?)
      found = true;
    } /* else {
      Serial.println("[verifyAck] No match");
      Serial.print("[verifyAck] ackPacket.sourceMAC(): ");
      Serial.println(ackPacket.sourceMAC(),HEX);
      Serial.print("[verifyAck] element.packet.destinationMAC(): ");
      Serial.println(element.packet.destinationMAC(),HEX);
      Serial.print("[verifyAck] ackPacket.sequenceNumber(): ");
      Serial.println(ackPacket.sequenceNumber());
      Serial.print("[verifyAck] element.packet.sequenceNumber(): ");
      Serial.println(element.packet.sequenceNumber());
    } */
    i++;
  }
  if ( !found ) {
    Serial.println("[verifyAck] ACK verification failed");
    ackPacket.serialOut();
  }
  return found;
}

void resendFunction() {
  Serial.println("[resendFunction] Timer fired...");
  int bufferSize = ackBuffer->size();
  if ( bufferSize > 0 ) {
    Serial.println("[resendFunction] Check Buffer...");
    checkAckBuffer = true;
    state = TX_State;
  } else {
    Serial.println("[resendFunction] Buffer empty");
    checkAckBuffer = false;
  }
  resendTimerActive = false;
}

void setSensorFlag () {
  Serial.println("[setSensorFlag] Timer fired...");
  readSensors = true;
  state = TX_State;
}

void pumpStateControl() {
  Serial.println("[pumpStateControl] Interrupt...");
  turnPumpOff = false;  // Not necessarily what is being requested...
  changePumpState = true;
  state = TX_State;
}

void pumpStateCheck() {
/*
  Or just:
*/
    changePumpState = true;
    turnPumpOff = true;
    pumpTimerActive = false;
    state = TX_State;
/*
  Also, check that our change request will get serviced if we are initally bumped by a higher priority TX action (i.e. we'll 
  get back into TX_State automatically, somehow...)
*/
/*
  unsigned long tickerTime = millis() - lastMessageTime;
  if ( tickerTime > wdtLimit ) {
    changePumpState = true;
    turnPumpOff = true;
    pumpTimerActive = false;
    state = TX_State;
  } else {
    TimerStart( &pumpWatchDogTimer );
    pumpTimerActive = true;
  }
*/
}

void readBatteryVoltage () {
  uint16_t batteryVoltage = getBatteryVoltage();
  Serial.print("[readBatteryVoltage] Battery Voltage: ");
  Serial.print( batteryVoltage );
  Serial.println(" mV");
  outPacket.setPacketType(VOLTAGE);
  outPacket.setVoltage(batteryVoltage);
}

void readBMESensor () {
  Serial.println("[readBMESensor] Read sensor...");
/*
 * If this node is reading atmospheric conditions from a BME sensor
 * All values recorded as integers, temperature multiplied by 10 to keep one decimal place
*/
  digitalWrite(Vext, LOW);    // Turn on the external power supply
  delay(50);                  // Give everything a moment to settle down
  Wire.begin();               // On with the show...
  
  if (bme280.init(0x76)) {
    Serial.println( "[readBMESensor] BME sensor [0x76] initialisation at complete" );
  } else if (bme280.init(0x77)) {
    Serial.println( "[readBMESensor] BME sensor [0x77] initialisation at complete" );
  } else {
    Serial.println( "[readBMESensor] BME sensor initialisation error" );
  }

  delay(100); // The BME280 needs a moment to get itself together... (50ms is too little time)

  temperature = (int) (10*bme280.getTemperature());
  pressure = (int) (bme280.getPressure() / 91.79F);
  humidity = (int) (bme280.getHumidity());

  outPacket.setPacketType(ATMOSPHERE);
  outPacket.setTemperature(temperature);
  outPacket.setPressure(pressure);
  outPacket.setHumidity(humidity);

  Wire.end();

  Serial.println("");
  Serial.print("[readBMESensor] Temperature: ");
  Serial.println((float) temperature/10, 1);
  Serial.print("[readBMESensor] Pressure: ");
  Serial.println(pressure);
  Serial.print("[readBMESensor] Humidity: ");
  Serial.println(humidity);
}

void deactivateOled() {
  Serial.println("[deactivateOled] Timer fired, OLED Display deactivated");
  if ( oledOn ) {
    myDisplay.clear();
    myDisplay.displayOff();
  }
  oledOn = false;
}

void toggleDisplay() {
  if ( oledOn ) {
    myDisplay.clear();
    myDisplay.displayOff();
  } else {
    myDisplay.displayOn();
    TimerStart( &oledTimer );
  }
  oledOn = !oledOn;
}

void activateSoftRev() {
  softRev = true;
}

void displaySoftRev() {
  String sketchRev = String(sketchRevision.major) + "." + String(sketchRevision.minor) + "." + String(sketchRevision.minimus);

  Serial.println("[displaySoftwareRevison] Display software revision data...");
  Serial.println("[displaySoftwareRevison] Sketch   " + sketchRev);
  Serial.println("[displaySoftwareRevison] Packet   " + inPacket.softwareRevision(PACKET_HANDLER));
  Serial.println("[displaySoftwareRevison] Node     " + inPacket.softwareRevision(NODE_HANDLER));
  Serial.println();
  
  if ( oledOn ) {
    myDisplay.clear();
    myDisplay.setFont(ArialMT_Plain_10);
    myDisplay.setTextAlignment(TEXT_ALIGN_LEFT);
    myDisplay.drawString(5,6,"Sketch");
    myDisplay.drawString(60,6,sketchRev);
    myDisplay.drawString(5,18,"Packet");
    myDisplay.drawString(60,18,inPacket.softwareRevision(PACKET_HANDLER));
    myDisplay.drawString(5,30,"Node");
    myDisplay.drawString(60,30,inPacket.softwareRevision(NODE_HANDLER));
    myDisplay.display();
  }
  softRev = false;
}